"""
DataFrame Schemas with Pandera.
"""

# __all__ = ['*']
__author__ = "Fernando Aristizabal"

from typing import List, Optional

import pandas as pd
import pandera as pa
from pandera.typing import Series, Index, Int64

"""
# TODO:
- Define custom datatypes that are unions of base types?
    - several of the float types below could be set to numeric
    - the band column could be set to numeric or str to allow for datasets.
"""


class Xrspatial_crosstab_df(pa.DataFrameModel):  # pragma: no cover
    """Defines the schema for output of `xrspatial.zonal.crosstab()`"""

    zone: Series[float]
    # TODO: Set schema to include any number of columns.
    # columns: pa.Column(name=Any, dtype=float, nullable=True, checks=None)
    # columns: Series[float] = pa.Field(regex=r"^(0|[1-9][0-9]*)$")

    class Config:
        coerce = True
        strict = False


class Sample_identifiers(pa.DataFrameModel):  # pragma: no cover
    """Crosstab DF schema"""

    band: Series[str]
    # this is where we extend these identifiers to map or catalog level identifiers.
    # NOTE: Sample identifiers are currently columns. Should they be indices instead???
    idx: Index[Int64]

    class Config:
        coerce = True
        strict = True

    @classmethod
    def columns(cls) -> List[str]:
        """Gives access to columns from DataFrameModel"""
        return list(super().to_schema().columns.keys())


class Crosstab_df(Sample_identifiers):  # pragma: no cover
    """Crosstab DF schema

    Inherits columns from `Sample_identifiers`
    """

    candidate_values: Series[float]
    benchmark_values: Series[float]
    agreement_values: Optional[Series[float]]
    counts: Series[float]

    class Config:
        coerce = True
        strict = True


class Pivoted_crosstab_df(pa.DataFrameModel):  # pragma: no cover
    """Pivoted Crosstab DF schema"""

    row_idx: Index[Int64] = pa.Index(name="candidate_values")
    col_idx: Series[str]

    class Config:
        coerce = True
        strict = True

    @pa.dataframe_check
    def column_index_name(cls, df: pd.DataFrame) -> Series[bool]:
        """Checks that column index name is 'benchmark_values'"""
        return df.columns.name == "benchmark_values"


class Conditions_df(Sample_identifiers):  # pragma: no cover
    """Cateogrical conditions df

    Inherits columns from `Sample_identifiers`
    """

    tp: Optional[Series[float]]
    tn: Optional[Series[float]] = pa.Field(nullable=True)
    fp: Optional[Series[float]]
    fn: Optional[Series[float]]

    class Config:
        coerce = True
        ordered = False  # this is TEMP. Should set id's to indices?


class Metrics_df(Conditions_df):  # pragma: no cover
    """Metrics DF schema

    Inherits columns from `Conditions_df`
    """

    class Config:
        coerce = True
        strict = False  # set to False, bc columns could include any number of metrics



# Test.py

from gval.statistics.categorical_statistics import CategoricalStatistics as CatSt

# Import class and instantiate
cat_stats = CatSt()


@cat_stats.register_function(name="test")
def test(tp: float, fn: int) -> float:
    return tp + fn


@cat_stats.register_function_class(vectorize_func=True)
class Tester:
    @staticmethod
    def test5(tp: int, fn: int) -> float:
        return tp + fn

    @staticmethod
    def test6(tp: int, fn: int) -> float:
        return tp + fn


def register(stat_obj):
    @stat_obj.register_function(name="test")
    def test(tp: float, fn: int) -> float:
        return tp + fn

    @stat_obj.register_function_class(vectorize_func=True)
    class Tester:
        @staticmethod
        def test5(tp: int, fn: int) -> float:
            return tp + fn

        @staticmethod
        def test6(tp: int, fn: int) -> float:
            return tp + fn



"""
Computes continuous value metrics given an agreement map.
"""

# __all__ = ['*']
__author__ = "Fernando Aristizabal"

from typing import Iterable, Union

import pandera as pa
import pandas as pd
from pandera.typing import DataFrame
import xarray as xr

from gval import ContStats
from gval.utils.schemas import Metrics_df


@pa.check_types
def _compute_continuous_metrics(
    agreement_map: Union[xr.DataArray, xr.Dataset],
    candidate_map: Union[xr.DataArray, xr.Dataset],
    benchmark_map: Union[xr.DataArray, xr.Dataset],
    metrics: Union[str, Iterable[str]] = "all",
) -> DataFrame[Metrics_df]:
    """
    Computes categorical metrics from a crosstab df.

    Parameters
    ----------
    agreement_map : Union[xr.DataArray, xr.Dataset]
        Agreement map, error based (candidate - benchmark).
    candidate_map : Union[xr.DataArray, xr.Dataset]
        Candidate map.
    benchmark_map : Union[xr.DataArray, xr.Dataset]
        Benchmark map.
    metrics : Union[str, Iterable[str]], default = "all"
        String or list of strings representing metrics to compute.

    Returns
    -------
    DataFrame[Metrics_df]
        Metrics DF with computed metrics per sample.

    Raises
    ------
    ValueError
        If metrics is not a string or list of strings.

    References
    ----------
    .. [1] [7th International Verification Methods Workshop](https://www.cawcr.gov.au/projects/verification/)
    .. [2] [3.3. Metrics and scoring: quantifying the quality of predictions](https://scikit-learn.org/stable/modules/model_evaluation.html)
    """

    # compute error based metrics such as MAE, MSE, RMSE, etc. from agreement map and produce metrics_df
    statistics, names = ContStats.process_statistics(
        metrics,
        error=agreement_map,
        candidate_map=candidate_map,
        benchmark_map=benchmark_map,
    )

    # create metrics_df
    metric_df = dict()
    for name, stat in zip(names, statistics):
        metric_df[name] = stat

    def is_nested_dict(d):
        # if not isinstance(d, dict):
        #     return False
        return any(isinstance(v, dict) for v in d.values())

    if is_nested_dict(metric_df):
        metric_df = pd.DataFrame.from_dict(metric_df, orient="index").transpose()
        metric_df.reset_index(inplace=True)
        metric_df.rename(columns={"index": "band"}, inplace=True)
        metric_df["band"] = metric_df["band"].str.replace("band_", "")

    else:
        # dataarray
        metric_df = pd.DataFrame(metric_df, index=[0])

        # add band
        metric_df.insert(0, "band", "1")

    return metric_df



from numbers import Number
from typing import Union, Iterable, Optional

import pandas as pd
from pandera.typing import DataFrame

from gval.comparison.compute_categorical_metrics import _compute_categorical_metrics
from gval.utils.schemas import Metrics_df


@pd.api.extensions.register_dataframe_accessor("gval")
class GVALDataFrame:
    """
    Class for extending pandas DataFrame functionality

    Attributes
    ----------
    _obj : pd.DataFrame
        Object to use off the accessor
    """

    def __init__(self, pandas_obj):
        self._obj = pandas_obj

    def compute_categorical_metrics(
        self,
        positive_categories: Union[Number, Iterable[Number]],
        negative_categories: Union[Number, Iterable[Number]],
        metrics: Union[str, Iterable[str]] = "all",
        average: str = "micro",
        weights: Optional[Iterable[Number]] = None,
    ) -> DataFrame[Metrics_df]:
        """
        Computes categorical metrics from a crosstab df.

        Parameters
        ----------
        crosstab_df : DataFrame[Crosstab_df]
            Crosstab DataFrame with candidate, benchmark, and agreement values as well as the counts for each occurrence.
        positive_categories : Optional[Union[Number, Iterable[Number]]]
            Number or list of numbers representing the values to consider as the positive condition. For average types "macro" and "weighted", this represents the categories to compute metrics for.
        negative_categories : Optional[Union[Number, Iterable[Number]]], default = None
            Number or list of numbers representing the values to consider as the negative condition. This should be set to None when no negative categories are used or when the average type is "macro" or "weighted".
        metrics : Union[str, Iterable[str]], default = "all"
            String or list of strings representing metrics to compute.
        average : str, default = "micro"
            Type of average to use when computing metrics. Options are "micro", "macro", and "weighted".
            Micro weighing computes the conditions, tp, tn, fp, and fn, for each category and then sums them.
            Macro weighing computes the metrics for each category then averages them.
            Weighted average computes the metrics for each category then averages them weighted by the number of weights argument in each category.
        weights : Optional[Iterable[Number]], default = None
            Weights to use when computing weighted average. Elements correspond to positive categories in order.

            Example:

            `positive_categories = [1, 2]; weights = [0.25, 0.75]`

        Returns
        -------
        DataFrame[Metrics_df]
            Metrics DF with computed metrics per sample.

        Raises
        ------
        ValueError
            Value is shared in positive and negative categories.
        ValueError
            Category not found in crosstab df.
        ValueError
            Cannot use average type with only one positive category.
        ValueError
            Number of weights must be the same as the number of positive categories.
        ValueError
            Cannot use average type with negative_categories as not None. Set negative_categories to None for this average type.

        References
        ----------
        .. [1] [Evaluation of binary classifiers](https://en.wikipedia.org/wiki/Evaluation_of_binary_classifiers)
        .. [2] [7th International Verification Methods Workshop](https://www.cawcr.gov.au/projects/verification/#Contingency_table)
        .. [3] [3.3. Metrics and scoring: quantifying the quality of predictions](https://scikit-learn.org/stable/modules/model_evaluation.html)
        """

        return _compute_categorical_metrics(
            crosstab_df=self._obj,
            metrics=metrics,
            positive_categories=positive_categories,
            negative_categories=negative_categories,
            average=average,
            weights=weights,
        )



"""
Comparison functionality
 - Includes pairing and comparison functions
 - crosstabbing functionality
 - Would np.meshgrid lend itself to pairing problem?
TODO:
    - Have not tested parallel case.
    - How to handle xr.Datasets, multiple bands, and multiple variables.
    - consider making a function registry to store pairing functions and their names in a dictionary
        - [Guide to function registration in Python with decorators](https://blog.miguelgrinberg.com/post/the-ultimate-guide-to-python-decorators-part-i-function-registration)
"""

# __all__ = ['*']
__author__ = "Fernando Aristizabal"

from typing import Iterable, Optional, Union, Callable
from numbers import Number

import numpy as np
import pandas as pd
from xrspatial.zonal import crosstab
import xarray as xr
import pandera as pa
from pandera.typing import DataFrame
import dask

from gval.utils.schemas import Xrspatial_crosstab_df, Crosstab_df
from gval.utils.loading_datasets import _check_dask_array


@pa.check_types
def _convert_crosstab_to_contigency_table(
    crosstab_df: DataFrame[Xrspatial_crosstab_df],
    band_name: str,
    band_value: Union[str, Number],
) -> DataFrame[Crosstab_df]:
    """
    Reorganizes crosstab output to Crosstab 2D DataFrame format.

    ----------
    Parameters
    crosstab_df : DataFrame[Xrspatial_crosstab_df]
        Output DataFrame from :func:`xarray-spatial.zonal.crosstab`.

    Returns
    -------
    DataFrame[Crosstab_df]
        Crosstab DataFrame using candidate and benchmark conventions.
    """
    if isinstance(crosstab_df, dask.dataframe.core.DataFrame):
        crosstab_df = crosstab_df.compute()

    # renames zone, renames column index, melts dataframe, then resets the index.
    crosstab_df = (
        crosstab_df.rename(columns={"zone": "candidate_values"})
        .rename_axis(columns="benchmark_values")
        .melt(id_vars="candidate_values", value_name="counts", ignore_index=False)
        .reset_index(drop=True)
    )

    # add band column
    crosstab_df.insert(0, band_name, band_value)

    return crosstab_df


@pa.check_types
def _compute_agreement_values(
    crosstab_df: DataFrame[Crosstab_df],
    comparison_function: Callable[..., float],
) -> DataFrame[Crosstab_df]:
    """
    Computes agreement values from Crosstab DataFrame.

    Parameters
    ----------
    crosstab_df : DataFrame[Crosstab_df]
        Crosstab DataFrame.
    comparison_function : Callable[[float, float], float]
        Function to compute agreement values.

    Returns
    -------
    DataFrame[Crosstab_df]
        Crosstab DataFrame with agreement values.
    """

    def apply_pairing_function(row):
        return comparison_function(row["candidate_values"], row["benchmark_values"])

    # copy crosstab_df
    crosstab_df = crosstab_df.copy()

    agreement_values = crosstab_df.apply(apply_pairing_function, axis=1)

    crosstab_df.insert(3, "agreement_values", agreement_values)

    return crosstab_df


def _crosstab_docstring(dimension: Union[int, str], xarray_obj: str = "xr.DataArray"):
    """
    Docstring decorator for crosstab functions.

    Parameters
    ----------
    dimension : Union[int, str]
        Number of dimensions function will support. Use either 2, 3, or 2/3.
    xarray_obj : str, default = "xr.DataArray"
        Type of xarray object function accepts. xr.DataArray or xr.Dataset.

    Returns
    -------
    Callable
        Decorated crosstab_* function with new docstring.

    Raises
    ------
    ValueError
        "Pass 2, 3, or 2/3 for dimension argument."
    """

    def decorator(func):
        docstring = f"""
            Crosstab {dimension}-dimensional {xarray_obj} to produce Crosstab DataFrame.

            Parameters
            ----------
            candidate_map : {xarray_obj}
                Candidate map, {dimension}-dimensional.
            benchmark_map : {xarray_obj}
                Benchmark map, {dimension}-dimensional.
            allow_candidate_values : Optional[Iterable[Union[int,float]]], default = None
                Sequence of values in candidate to include in crosstab. Remaining values are excluded.
            allow_benchmark_values : Optional[Iterable[Union[int,float]]], default = None
                Sequence of values in benchmark to include in crosstab. Remaining values are excluded.
            exclude_value : Optional[Number], default = None
                Value to exclude from crosstab. This could be used to denote a no data value if masking wasn't used. By default, NaNs are not cross-tabulated.
            comparison_function : Callable[[float, float], float], default = None
                Function to compute agreement values. If None, then no agreement values are computed.

            Returns
            -------
            DataFrame[Crosstab_df]
                Crosstab DataFrame.

            References
            ----------
            .. [1] :func:[`xrspatial.zonal.crosstab`](https://xarray-spatial.org/reference/_autosummary/xrspatial.zonal.crosstab.html)
            .. [2] [xarray.rio._check_dimensions()](https://github.com/corteva/rioxarray/blob/9d5975624fa93b76c451457a97b342ba37dfc792/rioxarray/rioxarray.py)
            .. [3] [xr.rio._obj.dims](https://github.com/corteva/rioxarray/blob/9d5975624fa93b76c451457a97b342ba37dfc792/rioxarray/raster_array.py)
            """
        func.__doc__ = docstring.format(dimension, xarray_obj)
        return func

    return decorator


@pa.check_types
@_crosstab_docstring(2, "xr.DataArray")
def _crosstab_2d_DataArrays(
    candidate_map: xr.DataArray,
    benchmark_map: xr.DataArray,
    band_name: str = "band",
    band_value: Union[str, Number] = 1,
    allow_candidate_values: Optional[Iterable[Number]] = None,
    allow_benchmark_values: Optional[Iterable[Number]] = None,
    exclude_value: Optional[Number] = None,
    comparison_function: Optional[Callable[..., float]] = None,
) -> DataFrame[Crosstab_df]:
    """Please see `_crosstab_docstring` function decorator for docstring"""

    if _check_dask_array(candidate_map):
        candidate_map = candidate_map.drop("spatial_ref")
        benchmark_map = benchmark_map.drop("spatial_ref")

    crosstab_df = crosstab(
        zones=candidate_map,
        values=benchmark_map,
        zone_ids=allow_candidate_values,
        cat_ids=allow_benchmark_values,
        nodata_values=exclude_value,
    )

    # reorganize df to follow contingency table schema instead of xarray-spatial conventions
    crosstab_df = _convert_crosstab_to_contigency_table(
        crosstab_df, band_name, band_value
    )

    # insert agreement values
    if comparison_function is not None:
        crosstab_df = _compute_agreement_values(crosstab_df, comparison_function)

    return crosstab_df


@pa.check_types
@_crosstab_docstring(3, "xr.DataArray")
def _crosstab_3d_DataArrays(
    candidate_map: xr.DataArray,
    benchmark_map: xr.DataArray,
    allow_candidate_values: Optional[Iterable[Number]] = None,
    allow_benchmark_values: Optional[Iterable[Number]] = None,
    exclude_value: Optional[Number] = None,
    comparison_function: Optional[Callable[..., float]] = None,
) -> DataFrame[Crosstab_df]:
    """Please see `_crosstab_docstring` function decorator for docstring"""

    # check number of dimensions
    assert (
        len(candidate_map.shape) == len(benchmark_map.shape) == 3
    ), "Candidate and benchmark must both be 3-dimensional"

    # check dimensionality
    # Is this necessary or is this done with functionality that checks band coordinates and within crosstab()
    # assert (
    #    candidate_map.shape == benchmark_map.shape
    # ), f"Dimensionalities of candidate {candidate_map.shape} and benchmark {benchmark_map.#shape} must match."

    """
    NOTE:
    - Use of `xr.rio._obj.dims` or `xr.DataArray.rio._check_dimensions()` to get band name (sometimes called band)
        - is it always called band? may not always be band?
        - are there any other methods for doing this??
        - should this go elsewhere as it might be repeated?
        - Is this necessary? Can we remove? Under what circumstance would it not be "band"
    - xr.DataArray.rio._check_dimensions() only supports 2/3D arrays and throws this error:
        - *** rioxarray.exceptions.TooManyDimensions: Only 2D and 3D data arrays supported.
        - This is ok and is useful to restrict dimensionality.
    """
    # get band name
    band_name_candidate = candidate_map.rio._check_dimensions()
    band_name_benchmark = benchmark_map.rio._check_dimensions()

    # get coordinates for bands
    candidate_map_band_coordinates = candidate_map[band_name_candidate].values
    benchmark_map_band_coordinates = benchmark_map[band_name_benchmark].values

    # check coordinates of extra dim to make sure they are the same
    # TODO: Having dataarrays with different extra dim coords remains untested.
    np.testing.assert_equal(
        candidate_map_band_coordinates,
        benchmark_map_band_coordinates,
        f"Coordinates of candidate dimension, {band_name_candidate}, and benchmark, {band_name_benchmark}, must be the same.",
    )

    # cycle through extra dim
    previous_crosstab_df = None  # initializing to avoid having unset
    for i, b in enumerate(candidate_map_band_coordinates):
        crosstab_df = _crosstab_2d_DataArrays(
            candidate_map=candidate_map.sel({band_name_candidate: b}),
            benchmark_map=benchmark_map.sel({band_name_benchmark: b}),
            band_name=band_name_candidate,
            band_value=b,
            allow_candidate_values=allow_candidate_values,
            allow_benchmark_values=allow_benchmark_values,
            exclude_value=exclude_value,
            comparison_function=comparison_function,
        )

        # concats crosstab_dfs across bands
        if i > 0:
            crosstab_df = pd.concat(
                [previous_crosstab_df, crosstab_df], ignore_index=True
            )

        # save df for previous band
        previous_crosstab_df = crosstab_df

    return crosstab_df


@pa.check_types
@_crosstab_docstring("2/3", "xr.DataArray")
def _crosstab_DataArrays(
    candidate_map: xr.DataArray,
    benchmark_map: xr.DataArray,
    allow_candidate_values: Optional[Iterable[Number]] = None,
    allow_benchmark_values: Optional[Iterable[Number]] = None,
    exclude_value: Optional[Number] = None,
    comparison_function: Optional[Callable[..., float]] = None,
) -> DataFrame[Crosstab_df]:
    """Please see `_crosstab_docstring` function decorator for docstring"""

    # TODO: these can be predicates and optional exception raising
    # 3d
    if len(candidate_map.shape) == len(benchmark_map.shape) == 3:
        crosstab_func = _crosstab_3d_DataArrays
    # 2d
    elif len(candidate_map.shape) == len(benchmark_map.shape) == 2:
        crosstab_func = _crosstab_2d_DataArrays
    else:
        raise ValueError(
            "Candidate and benchmark must be both 2 or 3 dimensional only."
        )

    return crosstab_func(
        candidate_map=candidate_map,
        benchmark_map=benchmark_map,
        allow_candidate_values=allow_candidate_values,
        allow_benchmark_values=allow_benchmark_values,
        exclude_value=exclude_value,
        comparison_function=comparison_function,
    )


@pa.check_types
@_crosstab_docstring("3", "xr.Dataset")
def _crosstab_Datasets(
    candidate_map: xr.Dataset,
    benchmark_map: xr.Dataset,
    allow_candidate_values: Optional[Iterable[Number]] = None,
    allow_benchmark_values: Optional[Iterable[Number]] = None,
    exclude_value: Optional[Number] = None,
    comparison_function: Optional[Callable[..., float]] = None,
) -> DataFrame[Crosstab_df]:
    """Please see `_crosstab_docstring` function decorator for docstring"""

    if _check_dask_array(candidate_map):
        # TODO:  Currently there is an issue open on xarray spatial regarding dask dataset useage in crosstab
        # https://github.com/makepath/xarray-spatial/issues/777
        candidate_map = candidate_map.compute()
        benchmark_map = benchmark_map.compute()

    # gets variable names
    candidate_variable_names = list(candidate_map.data_vars)
    benchmark_variable_names = list(benchmark_map.data_vars)

    # checks matching variable names
    # TODO: Is this desired? Should we just check for matching variable lengths?
    np.testing.assert_equal(
        candidate_variable_names,
        benchmark_variable_names,
        "Variable names must match for candidate and benchmark",
    )

    # loop variables
    previous_crosstab_df = None  # initializing to avoid having unset
    for i, b in enumerate(candidate_variable_names):
        crosstab_df = _crosstab_2d_DataArrays(
            candidate_map=candidate_map[b],
            benchmark_map=benchmark_map[b],
            band_value=b,
            allow_candidate_values=allow_candidate_values,
            allow_benchmark_values=allow_benchmark_values,
            exclude_value=exclude_value,
            comparison_function=comparison_function,
        )

        # concats crosstab_dfs across bands
        if i > 0:
            crosstab_df = pd.concat(
                [previous_crosstab_df, crosstab_df], ignore_index=True
            )

        # save df for previous band
        previous_crosstab_df = crosstab_df

    return crosstab_df
